import os

import signal as sigos

import math

import numpy as np
import pandas as pd
from scipy import signal

from datetime import datetime

import logging
from foglamp.common import logger

import sp
import sp_vib


class SignalFunctions:

    FILTERS = ['highpass', 'lowpass', 'bandpass', 'stop', 'notch', 'resample', 'hilbert', 'denoise',
               'autocorrelate', 'detrend', 'normalize', 'wiener', 'savgol', 'comb', 'denoise_freq']

    def __init__(self, filter_list, sampling_rate, n_fft, logger):
        self.filter_list = filter_list
        self.sampling_rate = sampling_rate
        self.n_fft = n_fft
        self.nyq = self.sampling_rate / 2
        self.logger = logger

        self.logger.info("yes initialized")
        extracted_list_of_functions = []
        for filter_dict in self.filter_list['filters']:
            is_filter_available = True
            for filter_name, filter_config in filter_dict.items():
                if filter_name not in SignalFunctions.FILTERS:
                    self.logger.error("The filter {} is not supported".format(filter_name))
                    is_filter_available = False
                    break
                _fop = getattr(self, filter_name)
                if type(filter_config) == dict:
                    if filter_config == {}:
                        _f = (_fop, ())
                    else:
                        # assuming it contains only single key value pair
                        for _, value in filter_config.items():
                            _f = (_fop, (float(value),))
                elif type(filter_config) == list:
                    list_of_floats = []
                    for parameters in filter_config:
                        for _, value in parameters.items():
                            list_of_floats.append(float(value))
                    _f = (_fop, tuple(list_of_floats))

            if not is_filter_available:
                continue
            extracted_list_of_functions.append(_f)

        self.filter_list = extracted_list_of_functions

    def apply(self, data):

        # execute the filter(s) sequentially
        # we should perform fft along with time series operations
        # and store results differently.

        _fft_freq_original, _fft_amp_original = self.fft(data, self.n_fft)
        for f in self.filter_list:
            data = f[0](data, *f[1])
        _fft_freq, _fft_amp = self.fft(data, self.n_fft)
        return data, _fft_freq, _fft_amp, _fft_freq_original, _fft_amp_original

    def fft(self, ts, n_fft) -> tuple:  # tuple of freqs, amps
        n_fft = int(n_fft)
        return sp.fft(ts, fs=self.sampling_rate, N_fft=n_fft)

    def normalize(self, ser):
        """ normalize - normalize amplitude to zero """
        if type(ser) is tuple:
            # frequency
            _freq, _amp = ser
            _max = np.max(_amp)
            if _max == 0:
                return ser
            return (_freq, _amp / _max)

        # time
        _max = np.max(ser)
        return ser / _max

    def denoise_freq(self, ser):
        """ denoise_freq - get rid of chatter that will distract dnn's """
        NOISE_LIMIT = .001  # assume we're normalized, so we erase when less than 2 orders of magnitude
        if type(ser) is tuple:
            # frequency
            _freq, _amp = ser
            _idx = np.argsort(_amp)
            _amp = np.where(_amp > NOISE_LIMIT, _amp, 0)
            return (_freq, _amp)

        # time
        return ser

    def resample(self, ts, nsample):
        return sp.resample_y_to(ts, nsample)

    def highpass(self, ts, cutoff):
        cutoff /= self.nyq
        return sp.bandwidth_filter_high(ts, cutoff)

    def lowpass(self, ts, cutoff):
        cutoff /= self.nyq
        return sp.bandwidth_filter_low(ts, cutoff)

    def bandpass(self, ts, lower, upper):
        lower /= self.nyq
        upper /= self.nyq
        return sp.bandwidth_filter_bandpass(ts, [lower, upper])

    def bandstop(self, ts, lower, upper):
        lower /= self.nyq
        upper /= self.nyq
        return sp.bandwidth_filter_bandstop(ts, [lower, upper])

    def notch(self, ts, notch_f):
        w0 = notch_f / self.nyq
        if w0 >= 1 or w0 <= 0:
            raise ValueError("error: filter values must be (0,{})".format(self.nyq))
        Q = 30
        b, a = signal.iirnotch(w0, Q)
        return signal.filtfilt(b, a, ts, axis=0)

    def comb(self, ts, notch_f):
        self.logger.info("comb: f={} shape={}".format(notch_f, len(ts)))
        w0 = notch_f / self.nyq
        if w0 >= 1 or w0 <= 0:
            raise ValueError("error: filter values must be (0,{})".format(self.nyq))
        Q = 30
        b, a = signal.iircomb(w0, Q)
        return signal.filtfilt(b, a, ts)

    def hilbert(self, ts):
        return sp.hilbert(ts)

    def denoise(self, ts):
        return sp_vib.denoise(ts)

    def autocorrelate(self, ts):
        return sp_vib.autocorrelate(ts, self.sampling_rate)

    def detrend(self, ts):
        return signal.detrend(ts)

    def wiener(self, ts):
        return signal.wiener(ts)

    def savgol(self, ts):
        WIN_LEN = 101
        POLY_ORDER = 3
        return signal.savgol_filter(ts, WIN_LEN, POLY_ORDER)
