# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp-foglamp-documentation.readthedocs-hosted.com
# FOGLAMP_END

""" Module for signal filter plugin
    This filter executes one or more signal processing transformations on input data
"""

__author__ = "Douglas Orr"
__copyright__ = "Copyright (c) 2020 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

import os
import json

import logging
from foglamp.common import logger
from foglamp.plugins.common import utils
from foglamp.common.plugin_helpers import HandleMap, PluginHandle

import filter_ingest


_LOGGER = logger.setup(__name__, level=logging.DEBUG)

_module_dir = os.path.dirname(__file__)
_handle_map = HandleMap(__name__)

FOGLAMP_ROOT = os.getenv("FOGLAMP_ROOT", "/usr/local/foglamp")
FOGLAMP_DATA = os.getenv("FOGLAMP_DATA", os.path.join(FOGLAMP_ROOT, "data"))

DEFAULT_N_FFT = 8192
DEFAULT_SAMPLING_RATE = 8000


class SignalProHandle(PluginHandle):

    # mandatory parameters for PluginHandle which has an rpc back-end
    RPC_SERVER_NAME = 'sigfns_server'          # name of RPC server module
    RPC_CONFIG_MEMBERS = ['filter_list', 'output_dir',
                          'sampling_rate', 'n_fft'
    ]   # config members to send to rpc for its config

    def __init__(self, config, cbinfo):
        super().__init__(__name__, cbinfo)  # set up logger, callback

        # slots -- values will be set by config
        self.plugin = None
        self.enable = None
        self.input_asset = None
        self.output_asset = None
        self.forward_data = None
        self.sampling_rate = None
        self.sp_multi_filter = None
        self.data_points = None
        self.n_fft = None

        self.rpc = None
        self.input = []
        self.input_type = "ts" # xxx - may want to derive from data or configuration

        self.ts_column = None
        self.pick_data_points_from_readings = False
        self.reconfigure(config)  # first time through, creates and configures server

        self.shutdown_flag = False

    def reconfigure(self, new_config):
        """ reconfigure - make the current plugin state consistent with the new configuration """

        # our server state is pretty simple, just tell the plugin to re-form its filters
        self.config_update(new_config)
        self.rpc_setup(self._rpc_config(), _module_dir)

    def config_update(self, update):
        """ config_update - save config values as members in this handle """
        super().config_update(update)       # update the values in the handle

        self.FRAME_LEN = self.sampling_rate
        self.output_dir = os.path.join(FOGLAMP_DATA, self.output_asset)
        os.makedirs(self.output_dir, exist_ok=True)

        if self.sp_multi_filter == {}:
            raise ValueError("must specify spMultiFilter in JSON format.")

        self.filter_list = self.sp_multi_filter

        if len(self.data_points) > 0:
            # if data points are mentioned in plugin configuration
            self.data_points = self.data_points.split(',')
        else:
            # pick all the data points
            self.pick_data_points_from_readings = True

        self.logger.info("Plugin handle config update successful")
        self.shutdown_flag = False

    def ingest(self, original_data):
        """ ingest - main ingest entry point, send data frames to signal processing server """

        if self.shutdown_flag:
            return

        if self.rpc is None:
            return

        if len(original_data) == 0:
            _LOGGER.info("empty data received ")
            return

        # filter out the input the input asset
        data = list(filter(lambda d: (d['asset'] == self.input_asset), original_data))
        if self.pick_data_points_from_readings:
            self.data_points = list(data[0]['readings'].keys())
            self.pick_data_points_from_readings = False

        if not self.ts_column:
            self.ts_column = 'user_ts' if 'user_ts' in data[0] else (
                'timestamp' if 'timestamp' in data[0] else None)

        self.input.extend(data)

        if len(self.input) < self.FRAME_LEN:
            # more data is required to fill the buffer.
            return

        time_stamp_buffer = []
        for reading in self.input[:self.FRAME_LEN]:
            time_stamp_buffer.append(reading[self.ts_column])

        time_domain_results = []
        frequency_domain_frequencies = []
        frequency_domain_amplitudes = []
        original_fft_frequencies = []
        original_fft_amplitudes = []

        for datapoint in self.data_points:
            buffer = []
            for reading in self.input[:self.FRAME_LEN]:
                buffer.append(reading['readings'][datapoint])

            time_domain_result, frequency_domain_frequency, \
                frequency_domain_amplitude, original_fft_frequency, original_fft_amplitude = \
                self.rpc.process_data(buffer)

            time_domain_results.append(time_domain_result)
            frequency_domain_frequencies.append(frequency_domain_frequency)
            frequency_domain_amplitudes.append(frequency_domain_amplitude)
            original_fft_frequencies.append(original_fft_frequency)
            original_fft_amplitudes.append(original_fft_amplitude)

        # calculating the transpose the results
        time_domain_results = list(map(list, zip(*time_domain_results)))
        frequency_domain_amplitudes = list(map(list, zip(*frequency_domain_amplitudes)))
        frequency_domain_frequencies = list(map(list, zip(*frequency_domain_frequencies)))
        original_fft_amplitudes = list(map(list, zip(*original_fft_amplitudes)))
        original_fft_frequencies = list(map(list, zip(*original_fft_frequencies)))

        # we have calculated all the results
        # time domain results are pretty much straight forward to ingest
        time_output_readings = []
        for i, t in enumerate(time_domain_results):
            readings_dict = {}
            for j, datapoint in enumerate(self.data_points):
                readings_dict[datapoint] = time_domain_results[i][j]
            time_output_readings.append(
                {
                    'asset': self.output_asset + '_time_domain_result',
                    self.ts_column: time_stamp_buffer[i],
                    'readings': readings_dict,
                    'type': 'time_domain'
                }
            )
        # now frequency domain outputs need to be forwarded
        frequency_output_readings = []
        for i, t in enumerate(frequency_domain_frequencies):
            readings_dict = {}
            for j, datapoint in enumerate(self.data_points):
                readings_dict[datapoint + '_frequency'] = frequency_domain_frequencies[i][j]
                readings_dict[datapoint + '_amplitude'] = frequency_domain_amplitudes[i][j]
            frequency_output_readings.append(
                {
                    'asset': self.output_asset + '_frequency_domain_result',
                    # just the first time stamp is enough
                    self.ts_column: time_stamp_buffer[0],
                    'readings': readings_dict,
                    'type': 'frequency_domain'
                }
            )
        # now frequency domain outputs need to be forwarded for the original signal
        original_fft_readings = []
        for i, t in enumerate(original_fft_frequencies):
            readings_dict = {}
            for j, datapoint in enumerate(self.data_points):
                readings_dict[datapoint + '_frequency'] = original_fft_frequencies[i][j]
                readings_dict[datapoint + '_amplitude'] = original_fft_amplitudes[i][j]
            original_fft_readings.append(
                {
                    'asset': self.input_asset + '_frequency_domain_result',
                    # just the first time stamp is enough
                    self.ts_column: time_stamp_buffer[0],
                    'readings': readings_dict,
                    'type': 'frequency_domain'
                }
            )
        time_output_readings.extend(frequency_output_readings)
        time_output_readings.extend(original_fft_readings)

        if self.forward_data:
            time_output_readings.extend(original_data)
        filter_ingest.filter_ingest_callback(self.callback, self.ingest_ref, time_output_readings)

        self.input = self.input[self.FRAME_LEN:]

    def plugin_shutdown(self):
        self.logger.info("Plugin handle server shutdown.")
        self.shutdown_flag = True
        if self.rpc is not None:
            try:
                self.rpc.plugin_shutdown()
            except EOFError as e:
                pass
            self.logger.info("making none")
            self.rpc = None
            self.logger.info("made none")
        self.logger.info("Plugin handle server shutdown complete.")

    
#
#
### PLUGIN API ###
#
#

_DEFAULT_CONFIG = {
    'plugin': {
        'description': 'Signal processing filter plugin',
        'type': 'string',
        'default': 'sigfns',
        'readonly': 'true'
    },
    'enable': {
        'description': 'enable/disable this filter',
        'type': 'boolean',
        'default': 'true',
        'displayName': 'Enable plugin',
        'order': "1"
    },
    'inputAsset': {
        'description': 'Input asset name',
        'type': 'string',
        'default': 'timeseries',
        'displayName': 'Asset name of time series data',
        'order': "2"
    },
    'outputAsset': {
        'description': 'Name of asset where filtered signal is written',
        'type': 'string',
        'default': 'signalProcessedOutput',
        'displayName': 'Output data asset',
        'order': "3"
    },
    'dataPoints': {
        'description': 'List of data points (comma separated) to filter the data.',
        'type': 'string',
        'default': '',
        'displayName': 'Data points list',
        'order': "4"
    },
    'forwardData': {
        'description': 'Forward input data through filter',
        'type': 'boolean',
        'default': 'false',
        'displayName': 'Forward input data',
        'order': "5"
    },
    'samplingRate': {
        'description': 'Sampling rate of acceleration data',
        'type': 'integer',
        'default': str(DEFAULT_SAMPLING_RATE),
        'displayName': 'Sampling rate (samples per sec)',
        'order': "6"
    },
    'nFft': {
        'description': 'Default number of FFT lines for fft operations',
        'type': 'integer',
        'default': str(DEFAULT_N_FFT),
        'displayName': 'Number of FFT lines',
        'order': "7"
    },
    'spMultiFilter': {
        'description': 'Comma-separated list of filter names and arguments',
        'type': 'JSON',
        'default': '{}',
        'displayName': 'One or more filter specs (comma separated)',
        'order': "8"
    },
}

def plugin_info():
    """ Returns information about the plugin
    Args:
    Returns:
        dict: plugin information
    Raises:
    """
    _LOGGER.debug("plugin_info called")
    return {
        'name': 'sigfns',
        'version': '2.0.0',
        'mode': "none",
        'type': 'filter',
        'interface': '1.0',
        'config': _DEFAULT_CONFIG
    }

def plugin_init(config, ingest_ref, callback):
    """ Initialise the plugin
    Args:
        config: JSON configuration document for the Filter plugin configuration category
        ingest_ref:
        callback:
    Returns:
        data: JSON object to be used in future calls to the plugin
    Raises:
    """
    _handle = SignalProHandle(config, (ingest_ref, callback))
    _h = _handle_map.new_handle(_handle, config)
    
    return _h

def plugin_ingest(h, data):
    """ Absorb run signal processing algorithms on time series data "data"
    Args:
        handle: handle returned by the plugin initialisation call
        data: readings data
    """

    _handle = _handle_map.get_handle(h)
    _handle.ingest(data)

def plugin_reconfigure(h, new_config):
    """ Reconfigures the plugin

    Args:
        h: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    """
    _handle = _handle_map.get_handle(h)

    _handle.reconfigure(new_config)

    h['config'].update(new_config)

    return h

def plugin_shutdown(h):
    """ Shutdowns the plugin doing required cleanup.
    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
        plugin shutdown
    """
    _LOGGER.info('plugin_shutdown called')

    _handle = _handle_map.get_handle(h)
    _handle.plugin_shutdown()  # shutdown any RPC objects
    
    _handle_map.del_handle(h)  # remove our mapping
    _LOGGER.info("plugin_shutdown completed.")
