# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp-foglamp-documentation.readthedocs-hosted.com
# FOGLAMP_END

""" Module for signal filter plugin
    This filter executes one or more signal processing transformations on input data
"""

__author__ = "Douglas Orr"
__copyright__ = "Copyright (c) 2020 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

import os

import signal as sigos

import math

import numpy as np
import pandas as pd
from scipy import signal

from datetime import datetime

import logging
from foglamp.common import logger
from foglamp.common.plugin_helpers import PluginRPCServer

import sp
import sp_vib

from sig_defs import SignalFunctions

FORMAT_TS="%Y-%m-%d %H:%M:%S.%f"


class SignalFnsServer(PluginRPCServer):
    """ process data frames applying signal processing filters to them """

    def __init__(self, service_name):
        super().__init__(service_name)  # set up logging, iprpc server

        # slots -- values will be assigned by config_update
        self.sampling_rate = None
        self.sp_multi_filter = None
        self.n_fft = None
        self.logger.setLevel(logging.DEBUG)
        self.valid = False

        self.signal_funcs_obj = None

    def plugin_init(self, config):
        # plugin_init -- apply the server params to the current handle

        # signal server only manages the filter_list parameter
        # apply config params to current handle
        self.config_update(config)
        self.logger.info("filter list: {}".format(self.filter_list))
        self.logger.info("type of filter is {}".format(type(self.filter_list)))
        self.signal_funcs_obj = SignalFunctions(self.filter_list, self.sampling_rate,
                                                self.n_fft, self.logger)
        # create the functional filter operation list, including optional parameters
        # self.filter_list = list(map(lambda x: _filter_op(x.split(':')), self.filter_list))
        self.valid = True

    def _process(self, data):
        """ process frame - receive one frame of one column of data; 
            convert using sp filters and return the converted result """

        data, _fft_freq_original, _fft_amp_original, \
        _fft_freq, _fft_amp  = self.signal_funcs_obj.apply(data)
        return data.tolist(), _fft_freq.tolist(), _fft_amp.tolist(), \
               _fft_freq_original.tolist(), _fft_amp_original.tolist()

    def process_data(self, data):
        """ process_data - take an array of raw data, process, and write out and return frame file """
        if not self.valid:
            self.logger.info("Plugin shutdown in process")
            return [], [], [], [], []
        _data = np.asarray(data)
        return self._process(_data)

    def process_frames(self, f_dir, fnames):
        """ process_frame - take a frame file of data, process, write out and return new frame file """
        _out_names = []

        if self.data_column == '':
            raise ValueError("must specify dataColumn to extract frame data")

        for fname in fnames:
            _in_path = os.path.join(f_dir, fname)
            if not _in_path.endswith(".tspkl"):
                raise ValueError("{}: invalid frame type".format(_in_path))

            _frame = pd.read_pickle(_in_path)

            if self.data_column not in list(_frame.columns):
                raise ValueError("dataColumn '{}' not found in dataframe columns {}".format(self.data_column, list(_frame.columns)))

            _data = _frame[self.data_column].to_numpy()
            _data = self._process(_data)

            # make us a new frame
            _out_names.append(self._flush(_data))
            
        return _out_names

    def _flush(self, data):
        """ flush - create a pickle file with this frame of transformed data in it """
        def _gen_filename(pfx, ts, *args, suff="tspkl"):
            """ gen_filename - generate a new, unique file name """
            _seq = 0
            while True:
                _fname = "{}_{}_{}".format(pfx, ts.strftime(FORMAT_TS), _seq)
                for a in args: _fname += "_{}".format(a)
                _fname += "."+suff
                if not os.path.exists(os.path.join(self.output_dir, _fname)):
                    break
                _seq += 1
            return _fname

        if self.result_type == "freq":
            # end result is to create an fft file
            _fname = _gen_filename("FFT", datetime.utcnow(), len(data), suff="fftpkl")
            _freq, _amp = data   # data is a tuple
            data = {"freq": _freq, "amp": _amp}
        else:
            # timeseries is the default
            _fname = _gen_filename("TS", datetime.utcnow(), len(data))
            # assign an arbitrary data column label
            data = {"y": data}
            
        _out_path = os.path.join(self.output_dir, _fname)

        _frame = pd.DataFrame(data)
        _frame.to_pickle(_out_path)

        return _fname
        

    def plugin_shutdown(self):
        self.logger.info("Signal processing server shutdown")
        self.valid = False
        self.logger.info("Signal processing server Shutdown complete")
        raise SystemExit
    #
    ###  filter operations  ###
    #



server = SignalFnsServer('sp-fns')
server.serve()
